import re

def get_pcie_id(devname: str):
    """
    Gets the PCIe ID from the device name, under
    the assumption that the relation is as:
    - enp7s0 => 0000:07:00.0
    - eth1 => 0000:01:00.0
    """
    match = re.search(r'(\d+)', devname)
    if match:
        return f"0000:{int(match.group(1))}:00.0"
    return None

def analyze_pcaps(pcapA, pcaps, output_str, host):
    command = f"python3 utilities/pcap_diff.py {pcapA} working/{output_str} -i {pcaps[0]}"
    for i in range(1, len(pcaps)):
        command += f" pcaps[i]"
    print("Analyzing pcaps...", end="")
    host.execute(command, quiet=True)
    print("Done")
    command = f"python3 utilities/grapher.py working/{output_str}_1.csv"
    for i in range(1, len(pcaps)):
        command += f" {output_str}_{i+1}.csv"
    host.execute(command, quiet=True)
    



from scapy.plist import PacketList
from scapy.utils import rdpcap
from scapy.compat import raw
import argparse
import pandas as pd
import enum
from tqdm import tqdm
from bisect import bisect_left

CREASE_PROT = 0x6587
    
class I(enum.IntEnum):
    TLR = 0
    TIME = 1
    IDX_DELTA = 2
    TIME_DELTA = 3
    IPG = 4
    CUM_LAT = 5

# For a new value new_value, compute the new count, new mean, the new M2.
# mean accumulates the mean of the entire dataset
# M2 aggregates the squared distance from the mean
# count aggregates the number of samples seen so far
def update(existing_aggregate, new_value):
    (count, mean, m2) = existing_aggregate
    count += 1
    delta = new_value - mean
    mean += delta / count
    delta2 = new_value - mean
    m2 += delta * delta2
    return (count, mean, m2)

# Retrieve the mean, variance and sample variance from an aggregate
def finalize(existing_aggregate):
    (count, mean, m2) = existing_aggregate
    if count < 2:
        return (float("nan"), float("nan"), float("nan"))
    else:
        (mean, variance, sample_variance) = (mean, m2 / count, m2 / (count - 1))
        return (mean, variance, sample_variance)

def shortest_edit_sequence(perm1_raw, perm2_raw):
    perm1 = list(perm1_raw)
    perm2 = list(perm2_raw)
    pos_in_perm1 = {c: i for i, c in enumerate(perm1)}
    perm2_indices = [pos_in_perm1[c] for c in perm2]

    n = len(perm2_indices)
    pile_tops = []
    predecessors = [-1] * n
    pile_indices = [0] * n

    for i, x in enumerate(tqdm(perm2_indices, desc="Finding LIS in perm2_indices")):
        pos = bisect_left(pile_tops, x)
        if pos == len(pile_tops):
            pile_tops.append(x)
        else:
            pile_tops[pos] = x
        pile_indices[i] = pos
        if pos > 0:
            for j in range(i - 1, -1, -1):
                if pile_indices[j] == pos - 1 and perm2_indices[j] < x:
                    predecessors[i] = j
                    break

    lis_len = len(pile_tops)
    lis_indices = []
    k = pile_indices.index(lis_len - 1)
    while k != -1:
        lis_indices.append(k)
        k = predecessors[k]
    lis_indices = set(lis_indices)

    # Simulate moves on a working copy so indices are always correct
    index_in_perm1 = {c: i for i, c in enumerate(perm1)}
    orig_index_in_perm1 = index_in_perm1.copy()
    moves = []
    target_pos = 0
    i = 0
    prev_lis_char = None
    lis_offset = 0
    for i, c in enumerate(tqdm(perm2, desc="Finding first edit script offset")):
        if i in lis_indices:
            prev_lis_char = c
            break
    for i, c in enumerate(tqdm(perm2, desc="Building edit script")):
        if i in lis_indices:
            prev_lis_char = c
            lis_offset = 0
            continue
        source_pos = index_in_perm1[c]
        target_pos = index_in_perm1[prev_lis_char] + lis_offset
        if target_pos > source_pos:
            for j in range(source_pos + 1, min(target_pos + 1, len(perm1))):
                index_in_perm1[perm1[j]] -= 1
        else:
            target_pos += 1
            for j in range(source_pos - 1, max(target_pos - 1, -1), -1):
                index_in_perm1[perm1[j]] += 1
        lis_offset += 1
        index_in_perm1[c] = target_pos
        perm1.pop(source_pos)
        perm1.insert(target_pos, c)
        moves.append((c, source_pos, target_pos, orig_index_in_perm1[c]))
    return moves

def apply_edit_script(perm, edit_script):
    perm = list(perm)
    for _, from_idx, to_idx, _ in tqdm(edit_script, desc="Applying edit script"):
        elem = perm.pop(from_idx)
        perm.insert(to_idx, elem)
    return perm

def lin_dist(
        pktlist_a: PacketList,
        pktlist_b: PacketList,
        output_file: str,
        replayer: int = -1,
        starta: float = 0,
        startb: float = 0):
    a_indices = {}
    b_indices = {}
    agg_a0 = (0, 0, 0)
    agg_a1 = (0, 0, 0)
    agg_b0 = (0, 0, 0)
    agg_b1 = (0, 0, 0)
    m = len(pktlist_a)
    n = len(pktlist_b)
    r = {}
    q = {}
    ap = 0
    bp = 0
    a_common = []
    b_common = []
    u_a = []
    u_b = []
    a = []
    b = []
    hex0 = b'\x00\x00'.hex()
    hex1 = b'\x00\x01'.hex()
    for pkt in tqdm(pktlist_a, desc="Reading PCAP 1"):
        if int.from_bytes(raw(pkt)[-2:], 'big') == CREASE_PROT:
            tlr = raw(pkt)[-14:-6].hex()
            if replayer == 0 and tlr[-4:] == hex0:
                a.append((tlr, pkt.time))
            elif replayer == 1 and tlr[-4:] == hex1:
                a.append((tlr, pkt.time))
            elif replayer == -1:
                a.append((tlr, pkt.time))
    for pkt in tqdm(pktlist_b, desc="Reading PCAP 2"):
        if int.from_bytes(raw(pkt)[-2:], 'big') == CREASE_PROT:
            tlr = raw(pkt)[-14:-6].hex()
            if replayer == 0 and tlr[-4:] == hex0:
                b.append((tlr, pkt.time))
            elif replayer == 1 and tlr[-4:] == hex1:
                b.append((tlr, pkt.time))
            elif replayer == -1:
                b.append((tlr, pkt.time))
    for i, c in enumerate(tqdm(a, desc="First PCAP 1 Loop")):
        tlr = c[0]
        if starta == 0 and i == 0:
            starta = c[1]
        if tlr in a_indices:
            print(f"Warning: Duplicate TLR {tlr} found in a at index {i}, skipping")
            continue
        a_indices[tlr] = i
        if tlr[-4:] == hex0:
            agg_a0 = update(agg_a0, i)
        elif tlr[-4:] == hex1:
            agg_a1 = update(agg_a1, i)
    for i, c in enumerate(tqdm(b, desc="PCAP 2 Loop")):
        tlr = c[0]
        if startb == 0 and i == 0:
            startb = c[1]
        if tlr in b_indices:
            print(f"Warning: Duplicate TLR {tlr} found in b at index {i}, skipping")
            continue
        b_indices[tlr] = i
        if tlr in a_indices:
            r[tlr] = [0, bp]
            b_common.append([tlr, c[1], 0, 0, 0, 0])
            bp += 1
        else:
            u_b.append(tlr)
        if tlr[-4:] == hex0:
            agg_b0 = update(agg_b0, i)
        elif tlr[-4:] == hex1:
            agg_b1 = update(agg_b1, i)
    for i, c in enumerate(tqdm(a, desc="Second PCAP 1 Loop")):
        tlr = c[0]
        if tlr in b_indices:
            r[tlr][0] = ap
            bp = r[tlr][1]
            b_common[bp][I.IDX_DELTA] = ap - bp
            b_common[bp][I.TIME_DELTA] = c[1] - b_common[bp][I.TIME]
            if ap == 0 and bp == 0:
                b_common[bp][I.IPG] = 0
            elif ap == 0:
                b_common[bp][I.IPG] = 0 - (b_common[bp][I.TIME] - b_common[bp-1][I.TIME])
            elif bp == 0:
                b_common[bp][I.IPG] = (c[1] - a_common[ap-1][1])
            else:
                b_common[bp][I.IPG] = (c[1] - a_common[ap-1][1]) - (b_common[bp][I.TIME] - b_common[bp-1][I.TIME])
            b_common[bp][I.CUM_LAT] = (c[1] - starta) - (b_common[bp][I.TIME] - startb)
            ap += 1
            a_common.append([tlr, c[1]])
        else:
            u_a.append(tlr)

    b_tlrs = [item[I.TLR] for item in b_common]
    a_tlrs = [item[I.TLR] for item in a_common]
    moves = shortest_edit_sequence(b_tlrs, a_tlrs)
    # c = apply_edit_script(b_common, moves)
    # mismatches = 0
    # for i, item_c in enumerate(c):
    #     item_a = a_common[i]
    #     if item_c[0] != item_a[0]:
    #         mismatches += 1
    # print(f"Total mismatches found: {mismatches}")
    print(f"Total moves: {len(moves)}")
    for move in tqdm(moves, desc="Updating deltas"):
        b_common[move[3]][I.IDX_DELTA] = move[2] - move[1]

    print(f"Unique in A: {len(u_a)}/{m}")
    print(f"Unique in B: {len(u_b)}/{n}")
    print(f"Common: {len(a_common)}")
    print("Writing stats to txt file...", end="", flush=True)
    agg_a0 = finalize(agg_a0)
    agg_a1 = finalize(agg_a1)
    agg_b0 = finalize(agg_b0)
    agg_b1 = finalize(agg_b1)
    with open(output_file + ".txt", "w") as f:
        f.write(f"Unique in A: {len(u_a)}/{m}\n")
        f.write(f"Unique in B: {len(u_b)}/{n}\n")
        f.write(f"Common: {len(a_common)}\n")
        f.write(f"Total moves: {len(moves)}\n")
        # f.write(f"Total mismatches found: {mismatches}\n")
        f.write(f"Aggregate A0: {agg_a0}\n")
        f.write(f"Aggregate A1: {agg_a1}\n")
        f.write(f"Aggregate B0: {agg_b0}\n")
        f.write(f"Aggregate B1: {agg_b1}\n")
        for i in range(len(u_a)):
            f.write(f"Unique in A: {u_a[i]} | {a_indices[u_a[i]]}\n")
        for i in range(len(u_b)):
            f.write(f"Unique in B: {u_b[i]} | {b_indices[u_b[i]]}\n")

    print(f"Writing results to CSV...", end="", flush=True)
    df_overall = pd.DataFrame(b_common, columns=[I.TLR.name, I.TIME.name, I.IDX_DELTA.name, I.TIME_DELTA.name, I.IPG.name, I.CUM_LAT.name])
    df_overall.to_csv(output_file + ".csv", index=True)
    print("Done")

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Compare two PCAP files for consistency.")
    parser.add_argument("pcap1", type=str, help="Path to the first PCAP file")
    parser.add_argument("output", type=str, help="Name of the output file (without extension)")
    parser.add_argument("-i", nargs='+', type=str, help="Paths to the other PCAP file(s) to compare against the first one.")
    parser.add_argument("-r", type=int, default=-1, help="Replayer ID")
    parser.add_argument("-t", nargs='+', type=float, help="Start times")
    args = parser.parse_args()

    print(f"Reading packets from PCAP file 1...", end="", flush=True)
    skip1 = ["runsL2/run13.pcap"]
    a = rdpcap(args.pcap1)
    if args.pcap1 in skip1:
        print(f"Skipping a packet in {args.pcap1} as it is in the skip list.")
        a = a[2:]
    print(f"Done", flush=True)
    if not args.i or not args.i[0]:
        print("No second PCAP file provided, exiting.")
        exit(1)
    start1 = 0
    start2 = 0
    if args.t:
        start1 = args.t[0]
    for i in range(len(args.i)):
        pcap2 = args.i[i]
        print(f"Comparing with PCAP file {i + 2} ({pcap2})...")
        if not pcap2:
            print("No second PCAP file provided, exiting.")
            exit(1)
        print(f"Reading packets from PCAP file {i + 2}...", end="", flush=True)
        b = rdpcap(pcap2)
        print(f"Done", flush=True)
        if pcap2 in skip1:
            print(f"Skipping a packet in {pcap2} as it is in the skip list.")
            b = b[2:]
        start2 = args.t[i+1] if args.t else 0
        lin_dist(a, b, args.output+f"_{i + 1}", args.r, start1, start2)


from scapy.all import Ether, IPv6, send, Raw, sendp
import argparse
from datetime import datetime

TYPE_REPLAY_RECORD = 1
TYPE_REPLAY_CLEAR = 2
TYPE_REPLAY_RUN = 4
CREASE_PROT = 254
S_TO_NS = 1_000_000_000

def send_command(
        msg_type: int,
        iface: str,
        args: Raw
):
    if msg_type not in (TYPE_REPLAY_RECORD, TYPE_REPLAY_CLEAR, TYPE_REPLAY_RUN):
        raise ValueError(f"Invalid command type: {msg_type}")
    
    msg = Raw(bytes([msg_type]))/args
    pkts = []
    pkt = Ether(dst="ff:ff:ff:ff:ff:ff")/IPv6(dst="2001:abcd::3", src="2001:abcd::2", nh=CREASE_PROT)/msg
    pkts.append(pkt)

    for pkt in pkts:
        sendp(pkt, verbose=False, iface=iface)
    print(f"Commands sent")

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Send commands to Crease nodes")
    parser.add_argument("msg_type", type=int, choices=[TYPE_REPLAY_RECORD, TYPE_REPLAY_CLEAR, TYPE_REPLAY_RUN],
                        help="Type of command to send")
    parser.add_argument("-i", action="append", help="Interface to send out of")
    parser.add_argument("-a", action="append", help="Additional arguments to include in the command")

    args = parser.parse_args()
    msg_args = None
    curr_time = datetime.now().timestamp()

    if args.msg_type == TYPE_REPLAY_RECORD:
        if len(args.a) < 2:
            raise ValueError("For TYPE_REPLAY_RECORD, two arguments are required: <start offset> <end offset>")
        start_offset = float(args.a[0])
        end_offset = float(args.a[1])
        msg_args = Raw(int((curr_time + start_offset) * S_TO_NS).to_bytes(8, 'big'))/Raw(int((curr_time + end_offset) * S_TO_NS).to_bytes(8, 'big'))
    elif args.msg_type == TYPE_REPLAY_CLEAR:
        msg_args = Raw(b'\x00')
    elif args.msg_type == TYPE_REPLAY_RUN:
        if len(args.a) < 1:
            raise ValueError("For TYPE_REPLAY_RUN, one argument is required: <start offset>")
        start_offset = int(args.a[0])
        msg_args = Raw(int((curr_time + start_offset) * S_TO_NS).to_bytes(8, 'big'))

    print(f"Sending command type {args.msg_type} to receivers: {args.i}")
    send_command(args.msg_type, args.i[0], msg_args)


import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import argparse
import math

# #df = pd.read_csv("overall.csv")


# # counts, bins, patches = plt.hist(df["jitter"].abs(), weights=np.ones(len(df))/len(df), bins=100)
# # plt.xlabel('Value Range')
# # plt.ylabel('Probability (%)')

# # plt.plot(df["jitter"].multiply(1_000_000))
# # plt.xlabel("Packet Index")
# # plt.ylabel("Cumulative Latency [ns]")
# #print("Filtering")
# #filtered = df[df["ipg"] != 0]
# #print("Filtered")

# vals = []
# for item in df["ipg"]:
#     if item != 0:
#         vals.append(item*1_000_000)

# for item in vals:
#     if item == 0:
#         print("err")

# # bp = plt.boxplot(vals, patch_artist=True)
# # for patch in bp["boxes"]:
# #     patch.set(facecolor="cyan")
# # #plt.setp(bp["boxes"], color="blue")
# # plt.yscale('symlog')
# # plt.xticks([])
# # plt.ylabel("Packet Inter-Arrival Times [ns]")
# plt.hist(vals, weights=np.ones(len(vals))/len(vals), bins=100)
# plt.xlabel("IAT Deviation [ns]")
# plt.ylabel("Percentage")
# #plt.yscale('log')

def graph_cum_latency(dfs, main_colors, box_colors, color_offset, letters):
    
    run = 0
    for df in dfs:
        vals = df["CUM_LAT"].multiply(1_000).values
        plt.plot(vals, label=f"Run {letters[run]}", color=main_colors[run % len(main_colors) + color_offset])
        run += 1
    plt.xlabel("Packet Index")
    plt.ylabel("Cumulative Latency [us]")
    plt.legend()
    plt.show()

def graph_packet_dists(dfs, main_colors, box_colors, color_offset, letters):
    
    run = 0
    for df in dfs:
        vals = df["IDX_DELTA"].abs().values
        max_val = np.max(np.abs(vals))
        if max_val == 0:
            max_val = 1
        plt.hist(vals, weights=np.ones(len(vals))/len(vals), bins=max_val, alpha=0.5, label=f"Run {letters[run]}", color=main_colors[run % len(main_colors) + color_offset])
        run += 1
    plt.xlabel("Edit Script Distance")
    plt.ylabel("Percentage")
    plt.legend()
    plt.show()

def graph_packet_dists2(dfs, main_colors, box_colors, color_offset, letters):
    
    run = 0
    for df in dfs:
        vals = df["IDX_DELTA"].abs().values
        bp = plt.boxplot(vals, positions=[run+1], widths=0.5, patch_artist=True)
        plt.setp(bp["medians"], color="black")
        for patch in bp["boxes"]:
            patch.set(facecolor=box_colors[run % len(box_colors) + color_offset])
        run += 1
    plt.xlabel("Run")
    plt.ylabel("Distance")
    plt.xticks(range(1, run+1), [f"Run {letters[i]}" for i in range(run)])
    plt.legend()
    plt.show()

def graph_packet_dists3(dfs, main_colors, box_colors, color_offset, letters):
    
    run = 0
    for df in dfs:
        vals = df["IDX_DELTA"].abs().values
        plt.plot(vals, label=f"Run {letters[run]}", color=main_colors[run % len(main_colors) + color_offset])
        run += 1
    plt.xlabel("Packet Index")
    plt.ylabel("Edit Script Distance")
    plt.legend()
    plt.show()

def graph_packet_dists4(dfs, main_colors, box_colors, color_offset, letters):
    
    run = 0
    vals = {}
    for df in dfs:
        for item in df["IDX_DELTA"].abs().values:
            if item not in vals:
                vals[item] = 0
            vals[item] += 1
        plt.plot(vals.keys(), vals.values(), label=f"Run {letters[run]}", color=main_colors[run % len(main_colors) + color_offset], alpha=0.5)
        run += 1
    plt.xlabel("Edit Script Distance")
    plt.ylabel("Packets")
    plt.legend()
    plt.show()

def graph_packet_dists5(dfs, main_colors, box_colors, color_offset, letters):
    
    run = 0
    vals = {}
    ctr = 0
    for row in dfs[0].itertuples():
        x  = row.TLR
        y = abs(row.IDX_DELTA)
        if x not in vals:
            vals[x] = [0] * (len(dfs) + 1)
            vals[x][0] = ctr
        vals[x][run+1] = y
        ctr += 1
    run += 1
    for df in dfs[1:]:
        ctr = 0
        for row in df.itertuples():
            x  = row.TLR
            y = abs(row.IDX_DELTA)
            # if x not in vals:
            #     vals[x] = [0] * (len(dfs) + 1)
            #     vals[x][0] = ctr
            vals[x][run+1] = y
            ctr += 1
        run += 1
    data = []
    for key, value in vals.items():
        data.append(value)
    sorted_data = sorted(data, key=lambda x: x[0])
    for run in range(len(dfs)):
        plt.plot([d[run+1] for d in sorted_data], label=f"Run {letters[run]}", color=main_colors[run % len(main_colors) + color_offset], alpha=0.5)
    plt.xlabel("Edit Script Distance")
    plt.ylabel("Packets")
    plt.legend()
    plt.show()

def graph_iat_deviation(dfs, main_colors, box_colors, color_offset, letters):
    run = 0
    fig = plt.figure(figsize=(7, 7))
    plt.rcParams['font.size'] = 10
    ax = fig.add_subplot(projection='3d')
    for df in dfs:
        vals = df["IPG"].multiply(1_000_000_000).values
        pow_min = -5
        pow_max = 5
        xbins = []
        xcenter = []
        xwidth = []
        for i in range(pow_min, 0):
            for j in range(2):
                if j == 0:
                    binV = -10 ** abs(i)
                    xC = i - 1
                else:
                    binV = (-10 ** abs(i)) // 2
                    xC = i -0.5
                xW = 0.5
                xbins.append(binV)
                xcenter.append(xC)
                xwidth.append(xW)
        xbins.append(-1)
        xbins.append(0)
        xcenter.append(-1.25 / 2)
        xwidth.append(1.25)
        xbins.append(1)
        xcenter.append(1.25 / 2)
        xwidth.append(1.25)
        for i in range(1, pow_max + 1):
            for j in range(2):
                if j == 1:
                    binV = 10 ** abs(i)
                    xC = i + 1
                else:
                    binV = (10 ** abs(i)) // 2
                    xC = i + 0.5
                xW = 0.5
                xbins.append(binV)
                xcenter.append(xC)
                xwidth.append(xW)
        # print(xbins)
        histvals, edges = np.histogram(vals, bins=xbins, weights=np.ones(len(vals))/len(vals))
        # print(len(xcenter), len(histvals), len(xwidth))
        # for i in range(len(xcenter)):
        #     print(f"{xcenter[i]} {histvals[i]} {xwidth[i]} {xbins[i]} {xbins[i+1]}")
        ax.bar(xcenter, align='center', width=xwidth, height=histvals, zs=run, zdir="y", label=f"Run {letters[run]}", color=main_colors[run % len(main_colors) + color_offset], alpha=0.5)
        run += 1
    ax.set_xlabel("Inter-Arrival Time Deviation [ns]")
    ax.set_ylabel("Run")
    ax.set_zlabel("Percentage")
    ax.set_yticks(range(run))
    ax.set_yticklabels([f"{letters[i]}" for i in range(run)])
    ax.set_xticks([i for i in range(-6, 7)])
    #ax.set_zticks([i/10 for i in range(0, 11)])
    ax.axes.set_xlim3d(-6, 6)
    xticks = []
    for i in range(-5, 1):
        xticks.append(f"$-10^{abs(i)}$")
    xticks.append(0)
    for i in range(0, 6):
        xticks.append(f"$10^{i}$")
    # print(xticks)
    ax.set_xticklabels(xticks)
    plt.savefig("/home/ubuntu/output/iat_graph.png")

def graph_lat_deviation(dfs, main_colors, box_colors, color_offset, letters):
    run = 0
    fig = plt.figure(figsize=(7, 7))
    plt.rcParams['font.size'] = 10
    ax = fig.add_subplot(projection='3d')
    for df in dfs:
        vals = df["CUM_LAT"].multiply(1_000_000_000).values
        pow_min = -10
        pow_max = 10
        xbins = []
        xcenter = []
        xwidth = []
        for i in range(pow_min, 0):
            for j in range(2):
                if j == 0:
                    binV = -10 ** abs(i)
                    xC = i - 1
                else:
                    binV = (-10 ** abs(i)) // 2
                    xC = i -0.5
                xW = 0.5
                xbins.append(binV)
                xcenter.append(xC)
                xwidth.append(xW)
        xbins.append(-1)
        xbins.append(0)
        xcenter.append(-1.25 / 2)
        xwidth.append(1.25)
        xbins.append(1)
        xcenter.append(1.25 / 2)
        xwidth.append(1.25)
        for i in range(1, pow_max + 1):
            for j in range(2):
                if j == 1:
                    binV = 10 ** abs(i)
                    xC = i + 1
                else:
                    binV = (10 ** abs(i)) // 2
                    xC = i + 0.5
                xW = 0.5
                xbins.append(binV)
                xcenter.append(xC)
                xwidth.append(xW)
        # print(xbins)
        histvals, edges = np.histogram(vals, bins=xbins, weights=np.ones(len(vals))/len(vals))
        # print(len(xcenter), len(histvals), len(xwidth))
        # for i in range(len(xcenter)):
        #     print(f"{xcenter[i]} {histvals[i]} {xwidth[i]} {xbins[i]} {xbins[i+1]}")
        ax.bar(xcenter, align='center', width=xwidth, height=histvals, zs=run, zdir="y", label=f"Run {letters[run]}", color=main_colors[run % len(main_colors) + color_offset], alpha=0.5)
        run += 1
    ax.set_xlabel("Inter-Arrival Time Deviation [ns]")
    ax.set_ylabel("Run")
    ax.set_zlabel("Percentage")
    ax.set_yticks(range(run))
    ax.set_yticklabels([f"{letters[i]}" for i in range(run)])
    ax.set_xticks([i for i in range(-7, 8)])
    #ax.set_zticks([i/10 for i in range(0, 11)])
    ax.axes.set_xlim3d(-7, 7)
    xticks = []
    for i in range(-6, 1):
        xticks.append(f"$-10^{abs(i)}$")
    xticks.append(0)
    for i in range(0, 7):
        xticks.append(f"$10^{i}$")
    # print(xticks)
    ax.set_xticklabels(xticks)
    plt.savefig("/home/ubuntu/output/latency_graph.png")

def graph_iat_box(dfs, main_colors, box_colors, color_offset, letters):
    
    run = 0
    for df in dfs:
        vals = df["IPG"].multiply(1_000_000_000).values
        bp = plt.boxplot(vals, positions=[run+1], widths=0.5, patch_artist=True)
        plt.setp(bp["medians"], color="black")
        for patch in bp["boxes"]:
            patch.set(facecolor=box_colors[run % len(box_colors) + color_offset])
        run += 1
    plt.xlabel("Run")
    plt.ylabel("Inter-Arrival Time [ns]")
    plt.xticks(range(1, run+1), [f"{letters[i]}" for i in range(run)])
    plt.yscale('symlog')
    plt.show()

def metric_o(dfs, main_colors, box_colors, color_offset, letters):
    run = 0
    retvals = []
    for df in dfs:
        vals = []
        total = 0
        for item in df["IDX_DELTA"]:
            vals.append(abs(item))
            total += abs(item)
        print(f"Run {run+1} O: {total/((len(vals) * (len(vals) - 1)) / 2)}")
        retvals.append(total/((len(vals) * (len(vals) - 1)) / 2))
        run += 1
    return retvals

def metric_i(dfs, main_colors, box_colors, color_offset, letters):
    run = 0
    retvals = []
    for df in dfs:
        vals = []
        total = 0
        for item in df["IPG"]:
            vals.append(abs(item))
            total += abs(item)
        dt = df["TIME"].iloc[-1] - df["TIME"].iloc[0]
        print(f"Run {run+1} I: {total/(2 * dt)}")
        retvals.append(total/(2 * dt))
        run += 1
    return retvals

def metric_l(dfs, main_colors, box_colors, color_offset, letters):
    run = 0
    retvals = []
    for df in dfs:
        vals = []
        total = 0
        for item in df["CUM_LAT"]:
            vals.append(abs(item))
            total += abs(item)
        dt = df["TIME"].iloc[-1] - df["TIME"].iloc[0]
        print(f"Run {run+1} L: {total/(len(vals) * dt)}")
        retvals.append(total/(len(vals) * dt))
        run += 1
    return retvals

def graph_consistency(m_us, m_os, m_is, m_ls):
    fig = plt.figure()
    ax = fig.add_subplot(111, projection='3d')
    img = ax.scatter(m_is, m_os, m_ls, c=m_us, cmap='viridis', s=50)
    ax.set_ylabel('Ordering')
    ax.set_xlabel('Inter-Arrival Time')
    ax.set_zlabel('Latency')
    #ax.axes.set_xlim3d(0, 1)
    #ax.axes.set_ylim3d(0, 1)
    #ax.axes.set_zlim3d(0, 1)
    cbar = plt.colorbar(img, pad=0.1, shrink=0.5)
    cbar.set_label('Uniqueness')
    plt.show()

def distance_metrics(dfs):
    dms = []
    for df in dfs:
        dm = []
        avg = df["IDX_DELTA"].mean()
        abs_avg = df["IDX_DELTA"].abs().mean()
        dm.append(avg)
        dm.append(df["IDX_DELTA"].std())
        dm.append(abs_avg)
        dm.append(df["IDX_DELTA"].abs().std())
        dm.append(df["IDX_DELTA"].min())
        dm.append(df["IDX_DELTA"].abs().min())
        dm.append(df["IDX_DELTA"].max())
        dm.append(df["IDX_DELTA"].abs().max())
        dms.append(dm)

    return dms


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Visualize packet data from a CSV file.")
    parser.add_argument("-c", help="Color offset", type=int, default=0)
    parser.add_argument("-i", nargs='+', action="append", type=str, help="Path to the CSV files containing packet data.")
    args = parser.parse_args()
    dfs = []
    box_colors = ["lightblue", "lightorange", "lightgray", "lightpurple", "cyan", "lightpink", "olivedrab", "lightgray"]
    main_colors = ["blue", "orange", "gray", "purple", "cyan", "pink", "olive", "gray"]
    letters = "BCDEFGHIJKLMNOPQRSTUVWXYZ"

    for csv in args.i[0]:
        df = pd.read_csv(csv)
        dfs.append(df)
    print(f"Loaded {len(dfs)} CSV files.")
    plt.rcParams['figure.figsize'] = (7, 6)
    plt.rcParams['font.size'] = 12

    graph_iat_deviation(dfs, main_colors, box_colors, args.c, letters)
    graph_lat_deviation(dfs, main_colors, box_colors, args.c, letters)

    #m_us = [0] * len(dfs)
    m_us = [(1052594 + 1053824 - 2 * 1052594) / (1053824 + 1052594),
            (1053824 + 1053824 - 2 * 1053824) / (1053824 + 1053824),
            (1053586 + 1053824 - 2 * 1053586) / (1053824 + 1053586),
            (1053619 + 1053824 - 2 * 1053619) / (1053824 + 1053619)]
    m_us = [0] * len(dfs)
    for i in range(len(m_us)):
        print(f"Run {i+1} U: {m_us[i]}")
    m_os = metric_o(dfs, main_colors, box_colors, args.c, letters)
    m_is = metric_i(dfs, main_colors, box_colors, args.c, letters)
    m_ls = metric_l(dfs, main_colors, box_colors, args.c, letters)
    if m_os[0] > 0:
        dms = distance_metrics(dfs)
        print("Distance Metrics:")
        for dm in dms:
            print(f"Avg: {dm[0]:.2f}, Std: {dm[1]:.2f}, Abs Avg: {dm[2]:.2f}, Abs Std: {dm[3]:.2f}, Min: {dm[4]}, Abs Min: {dm[5]}, Max: {dm[6]}, Abs Max: {dm[7]}")
    for i in range(len(dfs)):
        k = 1 - math.sqrt(m_us[i] ** 2 + m_os[i] ** 2 + m_is[i] ** 2 + m_ls[i] ** 2) / 2
        print(f"Run {i+1} Consistency: {k * 100:.5f}%")

    for i in range(len(dfs)):
        row0 = dfs[i].iloc[0]
        rowN = dfs[i].iloc[-1]
        total_time = rowN["TIME"] - row0["TIME"]
        print(f"Run {i+1} total time: {total_time}s, packets: {len(dfs[i])}, pps: {len(dfs[i])/total_time}")


    run = 1
    for df in dfs:
        total = 0
        c_iat = 0
        c_lat = 0
        for val in df["IPG"]:
            total += 1
            if abs(val) < 0.00000001:
                c_iat += 1
        print(f"Run {run} IAT Within +- 10 ns: {c_iat/total * 100:.2f}%")
        run += 1
